/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.sensor.fritzmeier;

import au.com.bytecode.opencsv.CSVReader;
import au.com.bytecode.opencsv.bean.ColumnPositionMappingStrategy;
import au.com.bytecode.opencsv.bean.CsvToBean;
import com.vividsolutions.jts.geom.Coordinate;
import eu.agrosense.client.grid.Grid;
import eu.agrosense.client.grid.GridProvider;
import java.beans.PropertyEditor;
import java.beans.PropertyEditorManager;
import java.io.StringReader;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import nl.cloudfarming.client.logging.AppLogFactory;
import nl.cloudfarming.client.logging.AppLogger;
import nl.cloudfarming.client.model.SensorObservation;
import nl.cloudfarming.client.model.SensorMetadata;
import nl.cloudfarming.client.sensor.SensorPalette;
import nl.cloudfarming.client.sensor.SensorProcessor;
import org.netbeans.api.progress.ProgressHandle;
import org.netbeans.api.progress.ProgressHandleFactory;
import org.openide.filesystems.FileObject;
import org.openide.util.Lookup;
import org.openide.util.NbBundle;

/**
 * Fritzmeier File processor Processes Fritzmeier sensor files
 *
 * @author Gerben Feenstra
 */
public class FritzmeierFileProcessor extends SensorProcessor {

    // Messages
    private static final String LOGKEY_FAILED_TO_READ_FILE = NbBundle.getMessage(SensorProcessor.class, "sensor.common.fileprocessor.logerror.failed_to_read_file");
    private static final String LOGKEY_PROCESSING_SUMMARY_ERRORMSG = NbBundle.getMessage(SensorProcessor.class, "sensor.common.fileprocessor.logerror.nr_of_records_failed");
    private static final String LOGKEY_PROCESSING_START = NbBundle.getMessage(SensorProcessor.class, "sensor.common.fileprocessor.loginfo.process_start");
    private static final String LOGKEY_PROCESSING_SUMMARY_PROCESSEDMSG = NbBundle.getMessage(SensorProcessor.class, "sensor.common.fileprocessor.loginfo.nr_of_processed");
    private static final String LOGKEY_TYPE_DISPLAYNAME = NbBundle.getMessage(FritzmeierFileProcessor.class, "sensor.fritzmeier.fileprocessor.log.displayname");
    // Other variable declarations
    private final String delimiter = FritzmeierFileDescription.getDelimiter();
    private final ProgressHandle progressHandle;
    protected FritzmeierDataObject dataObject;
    PropertyEditor old;
    private static final String[] COLUMS = new String[]{"latitude", "longitude", "ddateStr", "utcTimeStr", "course", "latitudeOriginal", "longitudeOriginal", "altitude", "speed", "fixType", "irmi", "tcari", "osavi", "tcOs", "wdvi", "ndre"};
    private CsvToBean csv;
    private ColumnPositionMappingStrategy strat;
    private static final AppLogger LOG = AppLogFactory.getLogger(FritzmeierFileProcessor.class);

    public FritzmeierFileProcessor(FritzmeierDataObject toProcess) {
        // assert dataObject != null;
        LOG.entering("FritzmeierFileProcessor", "constructor");
        this.dataObject = toProcess;
        this.progressHandle = ProgressHandleFactory.createHandle(NbBundle.getMessage(SensorProcessor.class, "sensor.common.fileprocessor.loginfo.process_description", dataObject.getName()));
        old = PropertyEditorManager.findEditor(Date.class);
        strat = new ColumnPositionMappingStrategy();
        strat.setType(FritzmeierSensorData.class);
        strat.setColumnMapping(COLUMS);
        csv = new CsvToBean();
    }

    /**
     * Parses payload to FritzmeierSensorData
     *
     * @param payload Fritzmeier sensor line
     * @param dataFile
     * @return FritzmeierSensorData object
     * @throws ParseException
     * @throws FritzmeierParseException
     */
    public FritzmeierSensorData parse(String payload) throws FritzmeierParseException {
        try {
            // Use delimiter defined in FritzmeierFileDescription
            CSVReader csvReader = new CSVReader(new StringReader(payload), delimiter.charAt(0));
            List<FritzmeierSensorData> list = csv.parse(strat, csvReader);
            return (FritzmeierSensorData) list.get(0);
        } catch (RuntimeException ex) {
            LOG.error(ex.getMessage());
            throw new FritzmeierParseException(ex.getCause());
        }
    }

    public boolean isHeader(String line) {
        return line.startsWith("$");
    }

    @Override
    public Grid process() {
        List<SensorObservation> sensorData = new ArrayList<SensorObservation>();

        try {
            int success = 0;
            int progress = 0;
            int fail = 0;

            LOG.info(LOGKEY_PROCESSING_START, getLogDisplayName(), dataObject.getName());
            FileObject fileObject = dataObject.getPrimaryFile();

            try {
                this.progressHandle.start(Long.valueOf(fileObject.asLines().size()).intValue());

                for (String line : fileObject.asLines()) {
                    progressHandle.progress(++progress);
                    if (!isHeader(line)) {
                        try {
                            sensorData.add(parse(line));
                            success++;
                        } catch (Exception ex) {
                            fail++;
                        }
                    }
                }
                LOG.info(LOGKEY_PROCESSING_SUMMARY_PROCESSEDMSG, success);
                if (fail > 0) {
                    LOG.warning(LOGKEY_PROCESSING_SUMMARY_ERRORMSG, fail);
                }
            } finally {
                progressHandle.finish();
            }
        } catch (Exception ex) {
            LOG.error(LOGKEY_FAILED_TO_READ_FILE, dataObject.getName());
            LOG.trace(ex);
        }
        if(sensorData.isEmpty()) return null;

        double minx = sensorData.get(0).getLongitude();
        double maxx = sensorData.get(0).getLongitude();
        double miny = sensorData.get(0).getLatitude();
        double maxy = sensorData.get(0).getLatitude();

        for (SensorObservation sd : sensorData) {
            minx = (minx < sd.getLongitude()) ? minx : sd.getLongitude();
            miny = (miny < sd.getLatitude()) ? miny : sd.getLatitude();
            maxx = (maxx > sd.getLongitude()) ? maxx : sd.getLongitude();
            maxy = (maxy > sd.getLatitude()) ? maxy : sd.getLatitude();
        }

        GridProvider gridProvider = Lookup.getDefault().lookup(GridProvider.class);
        Grid grid = gridProvider.getTemporaryGridSection(new Coordinate(minx, miny), new Coordinate(maxx, maxy));

        for (SensorObservation sd : sensorData) {
            //TODO sample data gets casted into an int, so multiply with 1000 for example
            if (sd.getNdvi() != null) {
                grid.setRasterValue(new Coordinate(sd.getLongitude(), sd.getLatitude()), (int) (sd.getNdvi() * SensorPalette.SENSOR_DATA_MULTIPLIER));
            }

            if (sd.getWdvi() != null) {
                grid.setRasterValue(new Coordinate(sd.getLongitude(), sd.getLatitude()), (int) (sd.getWdvi() * SensorPalette.SENSOR_DATA_MULTIPLIER));
            }
        }
        return grid;
    }

    /**
     * Method which will create a SensorMetadata object from the preferences:
     * FritzmeierPreference.SENSOR_DATA_FILE_FORMAT
     * FritzmeierPreference.CROPCIRCLE_SENSOR_HEIGTH
     * FritzmeierPreference.NUMBER_OF_CROPCIRCLE_SENSORS
     * FritzmeierPreference.CROPCIRCLE_SENSOR_SPREAD
     * FritzmeierPreference.SENSOR_TYPE FritzmeierPreference.SPECTRAL_CHANNELS
     *
     * @return created SensorMetadata object
     */
    @Override
    protected SensorMetadata buildMetadataFromPreferences() {
        String fileFormat = FritzmeierPreference.SENSOR_DATA_FILE_FORMAT.getValue();
        int sensorHeigth = Integer.parseInt(FritzmeierPreference.FRITZMEIER_SENSOR_HEIGTH.getValue());
        int numberOfSensors = Integer.parseInt(FritzmeierPreference.NUMBER_OF_FRITZMEIER_SENSORS.getValue());
        String FritzmeierSensorSpread = FritzmeierPreference.FRITZMEIER_SENSOR_SPREAD.getValue();
        String sensorType = FritzmeierPreference.SENSOR_TYPE.getValue();
        String spectralChannels = FritzmeierPreference.SPECTRAL_CHANNELS.getValue();
        return new SensorMetadata(fileFormat, sensorHeigth, numberOfSensors, FritzmeierSensorSpread, sensorType, spectralChannels);
    }

    @Override
    protected String getFileDateFormat() {
        return FritzmeierPreference.SENSOR_DATA_FILE_FORMAT.getValue();
    }

    @Override
    protected String getLogDisplayName() {
        return LOGKEY_TYPE_DISPLAYNAME;
    }
}
